# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from chipscopy import CoreType
from chipscopy.api._detail.debug_core import DebugCore
from chipscopy.client.axis_trace_core_client import AxisTraceClient


class Trace(DebugCore["AxisTraceClient"]):
    def __init__(self, tcf_trace: AxisTraceClient):
        DebugCore.__init__(self, core_type=CoreType.AXIS_TRACE, core_tcf_node=tcf_trace)
        self.filter_by = {"uuid": self.core_info.uuid}
